# Generated by Django 4.2.10 on 2024-04-15 08:48

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0034_alter_chatmodeloptions_chat_model"),
    ]

    operations = [
        migrations.CreateModel(
            name="ProcessLock",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("name", models.CharField(choices=[("update_embeddings", "Update Embeddings")], max_length=200)),
                ("started_at", models.DateTimeField(auto_now_add=True)),
                ("max_duration_in_seconds", models.IntegerField(default=43200)),
            ],
            options={
                "abstract": False,
            },
        ),
    ]
